/*************************************************************************
* Simple TFTP server by J. Fustin
*
* This program was written as an assignment for Computer Networks course
* at FER (www.fer.hr/english). Its purpose is to be as simple as possible
* so everyone could learn how to perform basic communication with TFTP
* clients over UDP/IP.
*
* Change log:
* 01/2005 - initial release
* 03/2007 - refactored a bit
* 12/2009 - completely restructured
*************************************************************************/

#include "common.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <inttypes.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* Socket descriptor. */
int sck;

/* Client data. */
typedef struct
{
	unsigned int active;		/* 0 = inactive, 1 = active */
	
	struct sockaddr_in addr;	/* Client's address struct. */
	unsigned int addrlen;		/* Address struct length. */
	
	action act;					/* ACTION_UNKNOWN, ACTION_SEND or ACTION_RECEIVE */
	FILE* file;
	char filename[FILENAME_LEN];
	unsigned int readnext;
	
	char buffer[BUFF_SIZE];
	uint16_t blocknum;
	unsigned int retrynum;
	unsigned int finished;		/* 0 = in progress, 1 = finished */
} cl_data;

/* Array of clients. Max 32 simultaneous clients. */
cl_data cl[32];
unsigned int cl_count = 0;

/* Error messages. */
const char* errormsg[] =
{
	"Not defined, see error message (if any).",
	"File not found.",
	"Access violation.",
	"Disk full or allocation exceeded.",
	"Illegal TFTP operation.",
	"Unknown transfer ID.",
	"File already exists.",
	"No such user."
};

void print_instructions()
{
	const char* instructions =
		"Simple TFTP server by J. Fustin.\n"
		"Usage: server <port>\n";
	
	fprintf(stderr, "%s", instructions);
}

void fatal_error(const char* message)
{
	fprintf(stderr, "Fatal error: %s\n", message);
	exit(1);
}

void sigint_handler(int sig)
{
	printf("Bye bye.\n");
	exit(1);
}

/*************************************************************************
 * [in] packet, packetlen, client
 *************************************************************************/
void handle_receive(const char* packet, unsigned int packetlen,
					cl_data* client)
{
	int rd = 0;	/* number of bytes read from file */
	
	uint16_t opcode;
	uint16_t blocknum;
	
	parse_frame(packet, &opcode, &blocknum);
	
	switch(opcode)
	{
		case OPCODE_RRQ:
			/* This opcode is valid only for the first block. */
			if(client->blocknum != 1)
				break;
			
			/* See if the file exists. */
			if(!client->file)
			{
				client->file = fopen(packet + 2, "r");
				strncpy(client->filename, packet + 2, FILENAME_LEN - 1);
			}
			
			if(!client->file)
			{
				printf("File not found: %s\n", client->filename);
				client->active = 0;
				
				make_error_frame(1, errormsg[1], client->buffer);
				sendto(sck, client->buffer, 4 + strlen(errormsg[1]) + 1, 0,
					(struct sockaddr *) &client->addr, client->addrlen);
				
				return;
			}
			
			break;
			
		case OPCODE_ACK:
			/* Check if we got a correct ACK block. */
			if(blocknum != client->blocknum)
			{
				++client->retrynum;
			}
			else
			{
				client->readnext = 1;
				++client->blocknum;
				client->retrynum = 0;
				
				if(client->finished)
				{
					fclose(client->file);
					client->file = NULL;
					client->active = 0;
					printf("Sent successfully: %s\n", client->filename);
					return;
				}
			}
			
			break;
			
		case OPCODE_ERROR:
			/* Handle error. */
			printf("Client reported an error.\n");
			client->active = 0;
			return;
			break;
			
		default:
			/* Unexpected opcode. */
			printf("Unexpected opcode.\n");
			make_error_frame(4, errormsg[4], client->buffer);
			sendto(sck, client->buffer, 4 + strlen(errormsg[4]) + 1, 0,
				(struct sockaddr *) &client->addr, client->addrlen);
			return;
			break;
	}
	
	/* Send DATA frame. */
	if(client->readnext)
	{
		make_data_frame(client->blocknum, client->buffer);
		rd = fread(client->buffer + 4, 1, 512, client->file);
		
		if(rd < 512)
			client->finished = 1;
		
		client->readnext = 0;
	}
	
	sendto(sck, client->buffer, 4 + rd, 0,
		(struct sockaddr *) &client->addr, client->addrlen);
}

/*************************************************************************
 * [in] packet, packetlen, client
 *************************************************************************/
void handle_send(const char* packet, unsigned int packetlen,
				 cl_data* client)
{
	uint16_t opcode;
	uint16_t blocknum;
	
	parse_frame(packet, &opcode, &blocknum);
	
	switch(opcode)
	{
		case OPCODE_WRQ:
			/* This opcode is valid only for the zeroth block. */
			if(client->blocknum != 0)
				break;
			
			/* See if the file exists. */
			if(!client->file)
			{
				client->file = fopen(packet + 2, "w");
				strncpy(client->filename, packet + 2, FILENAME_LEN - 1);
			}
			
			if(!client->file)
			{
				printf("Cannot create file: %s\n", client->filename);
				client->active = 0;
				
				make_error_frame(2, errormsg[2], client->buffer);
				sendto(sck, client->buffer, 4 + strlen(errormsg[2]) + 1, 0,
					(struct sockaddr *) &client->addr, client->addrlen);
				
				return;
			}
			
			break;
		
		case OPCODE_DATA:
			/* Check if we got a correct DATA block. */
			if(blocknum != client->blocknum + 1)
			{
				++client->retrynum;
			}
			else
			{
				++client->blocknum;
				client->retrynum = 0;
				
				fwrite(packet + 4, 1, packetlen - 4, client->file);
				fflush(client->file);
				
				if(packetlen < 516)
				{
					fclose(client->file);
					client->file = NULL;
					client->finished = 1;
					client->active = 0;
					printf("Received successfully: %s\n", client->filename);
				}
			}
			
			break;
		
		case OPCODE_ERROR:
			/* Handle error. */
			printf("Client reported an error.\n");
			client->active = 0;
			return;
			break;
			
		default:
			/* Unexpected opcode. */
			printf("Unexpected opcode.\n");
			make_error_frame(4, errormsg[4], client->buffer);
			sendto(sck, client->buffer, 4 + strlen(errormsg[4]) + 1, 0,
				(struct sockaddr *) &client->addr, client->addrlen);
			return;
			break;
	}
	
	/* Send ACK frame. */
	make_ack_frame(client->blocknum, client->buffer);
	sendto(sck, client->buffer, 4, 0,
		(struct sockaddr *) &client->addr, client->addrlen);
}

/*************************************************************************
 * [in] packet, packetlen, client
 *************************************************************************/
void handle_packet(const char* packet, unsigned int packetlen,
				   cl_data* client)
{
	char buffer[BUFF_SIZE];
	uint16_t rcv_opcode;
	
	memcpy(buffer, packet, BUFF_SIZE);
	parse_frame(buffer, &rcv_opcode, NULL);
	
	switch(client->act)
	{
		case ACTION_RECEIVE:
			handle_receive(packet, packetlen, client);
			break;
		
		case ACTION_SEND:
			handle_send(packet, packetlen, client);
			break;
		
		case ACTION_UNKNOWN:
		{
			switch(rcv_opcode)
			{
				case OPCODE_RRQ:
					client->act = ACTION_RECEIVE;
					client->readnext = 1;
					client->blocknum = 1;
					handle_receive(packet, packetlen, client);
					break;
					
				case OPCODE_WRQ:
					client->act = ACTION_SEND;
					client->blocknum = 0;
					handle_send(packet, packetlen, client);
					break;
					
				case OPCODE_ERROR:
					/* Handle error. */
					printf("Client reported an error.\n");
					client->active = 0;
					break;
					
				default:
					/* Unexpected opcode. */
					break;
			}
		}
	}
}

void start_tftp_server(unsigned short port)
{
	int sel;	/* return value from select() */
	int rf;		/* return value from recvfrom() */
	
	unsigned int i;
	
	struct sockaddr_in srvaddr;
	struct sockaddr_in cliaddr;
	unsigned int cliaddrlen = sizeof(cliaddr);
	
	char buffer[BUFF_SIZE];
	
	/* Initialize host's address struct. */
	memset(&srvaddr, 0, sizeof(srvaddr));
	srvaddr.sin_family = AF_INET;
	srvaddr.sin_port = htons(port);
	srvaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	
	/* Create a connectionless (UDP/IP) socket. */
	sck = socket(AF_INET, SOCK_DGRAM, 0);
	
	if(sck == -1)
		fatal_error("Can not create socket.");
	
	/* Bind socket to specified address. */
	if(bind(sck, (struct sockaddr *) &srvaddr, sizeof(srvaddr)) == -1)
		fatal_error("Can not bind socket.");
	
	printf("TFTP server started on port %hd.\n", port);
	
	while(1)
	{
		sel = poll_socket(sck, 2);
		
		if(sel == -1 || sel == 0)
			continue;
		
		rf = recvfrom(sck, buffer, BUFF_SIZE, 0,
					(struct sockaddr *) &cliaddr, &cliaddrlen);
		
		/* See if we've already communicated with this client. */
		for(i = 0; i < cl_count; ++i)
		{
			if(cliaddr.sin_addr.s_addr == cl[i].addr.sin_addr.s_addr &&
				cliaddr.sin_port == cl[i].addr.sin_port)
			{
				break;
			}
		}
		
		/* Client not found. Add a new one. */
		if(i == cl_count)
		{
			cl[cl_count].active = 1;
			memcpy(&cl[cl_count].addr, &cliaddr, cliaddrlen);
			cl[cl_count].addrlen = cliaddrlen;
			cl[cl_count].act = ACTION_UNKNOWN;
			cl[cl_count].file = NULL;
			cl[cl_count].readnext = 0;
			cl[cl_count].blocknum = 0;
			cl[cl_count].retrynum = 0;
			cl[cl_count].finished = 0;
			++cl_count;
		}
		
		/* Dispatch packet to client. */
		handle_packet(buffer, rf, &cl[i]);
		
		/* Cleanup inactive clients. */
		for(i = 0; i < cl_count;)
		{
			if(!cl[i].active)
			{
				--cl_count;
				memcpy(&cl[i], &cl[cl_count], sizeof(cl_data));
			}
			else
			{
				++i;
			}
		}
	}
}

int main(int argc, char *argv[])
{
	unsigned short port;
	
	if(argc != 2)
	{
		print_instructions();
		exit(1);
	}
	
	signal(SIGINT, sigint_handler);
	
	port = atoi(argv[1]);
	start_tftp_server(port);
	
	return 0;
}

